#version 330 core

layout(location = 0) out vec4 fragColor;
layout(location = 1) out vec4 brightnessColor; //store brightness values for bloom effect

// Interpolated values from the vertex shaders
in vec2 uv_varyings;
in vec3 pos_worldSpace;
in vec3 normal_camSpace;
in vec3 eyeDir_camSpace;
in vec3 lightDir_camSpace;
in vec3 fireDir_camSpace;
in vec4 fireShadowCoord;
in vec4 lightShadowCoord;

// Values that stay constant for the whole mesh.
uniform sampler2D texture_color;
uniform vec3 lightPos_worldSpace;
uniform vec3 ambientLightColor;
uniform vec3 lightColor;
uniform vec3 fireColor;
uniform vec3 firePos_worldSpace;
uniform sampler2DShadow fireShadowMap;
uniform sampler2DShadow lightShadowMap;
uniform vec4 cameraShadowCoord;

#define stepSize 0.0015

//poissondisk
vec2 poissonDisk[16] = vec2[](
	vec2( -0.94201624, -0.39906216 ),
	vec2( 0.94558609, -0.76890725 ),
	vec2( -0.094184101, -0.92938870 ),
	vec2( 0.34495938, 0.29387760 ),
	vec2( -0.91588581, 0.45771432 ),
	vec2( -0.81544232, -0.87912464 ),
	vec2( -0.38277543, 0.27676845 ),
	vec2( 0.97484398, 0.75648379 ),
	vec2( 0.44323325, -0.97511554 ),
	vec2( 0.53742981, -0.47373420 ),
	vec2( -0.26496911, -0.41893023 ),
	vec2( 0.79197514, 0.19090188 ),
	vec2( -0.24188840, 0.99706507 ),
	vec2( -0.81409955, 0.91437590 ),
	vec2( 0.19984126, 0.78641367 ),
	vec2( 0.14383161, -0.14100790 )
);

// Returns a random number based on a vec3 and an int.
float random(vec3 seed, int i){
	vec4 seed4 = vec4(seed,i);
	float dot_product = dot(seed4, vec4(12.9898,78.233,45.164,94.673));
	return fract(sin(dot_product) * 43758.5453);
}

void main(){

	// constant Light emission properties
	float lightPower = 400.0f;
	float firePower = 200.0f;
	vec3 windowPos_worldSpace = vec3(-7.5f, 2.9f, -7.5f);
	
	
	// Material properties
	vec3 MaterialDiffuseColor = texture( texture_color, uv_varyings ).rgb;
	vec3 MaterialAmbientColor = ambientLightColor * MaterialDiffuseColor;
	vec3 MaterialSpecularColor = vec3(0.3,0.3,0.3);
	vec3 MaterialFireColor = fireColor * MaterialDiffuseColor;
	vec3 MaterialFireSpecularColor = MaterialFireColor * vec3(1.2f,1.2f,1.2f);
	

	//distance from window, lower bounded at 0.5
	float wind_dist = max(length( windowPos_worldSpace - pos_worldSpace), 0.5);

	// Distance to the light
	float lightDistance = length( lightPos_worldSpace - pos_worldSpace );

	// normalize normal
	vec3 n = normalize( normal_camSpace );
	// normalize light direction (from the fragment to the light)
	vec3 l = normalize( lightDir_camSpace );
	// Cosine of the angle between the normal and the light direction
	float cosTheta = clamp( dot( n,l ), 0,1 );
	
	// Eye vector (towards the camera)
	vec3 E = normalize(eyeDir_camSpace);
	// Direction in which the triangle reflects the light
	vec3 R = reflect(-l,n);

	// Cosine of the angle between the Eye vector and the Reflect vector
	float cosAlpha = clamp( dot( E,R ), 0,1 );

///////////////fire
	//distance to fire
	float fireDist = length( firePos_worldSpace - pos_worldSpace );
	// normalize fire direction (from the fragment to the fire)
	vec3 fireL = normalize( fireDir_camSpace );
	vec3 fireR = reflect(-fireL,n);
	float fireCosTheta = clamp( dot( n,fireL ), 0,1 );
	float fireCosAlpha = clamp( dot( E,fireR ), 0,1 );

/////////////////shadows
//---------1.shadow map for fire
	float fireVisibility = 1.0;
	float bias = 0.005*tan(acos(fireCosTheta));
	bias = clamp(bias, 0,0.01);
	//float bias = -0.01;

	if(fireShadowCoord.w >0.0)
	{
		
		vec4 shadowCoordProj = fireShadowCoord/fireShadowCoord.w;
		if(shadowCoordProj.x >= 0.0 && shadowCoordProj.x < 1.0 &&
			shadowCoordProj.y >= 0.0 && shadowCoordProj.y < 1.0 &&
			shadowCoordProj.z >= 0.0 && shadowCoordProj.z < 1.0)
		{
			if(textureProj( fireShadowMap, fireShadowCoord )  <  (shadowCoordProj.z-bias))
			{
				// Sample the shadow map 4 times
				for (int i=0;i<8;i++){
					//int index = i;
					int index = int(16.0*random(floor(pos_worldSpace.xyz*1000.0), i))%16;
					fireVisibility -= 0.125*(1.0-texture(fireShadowMap, vec3(shadowCoordProj.xy + poissonDisk[index]/1500.0 , (shadowCoordProj.z-bias)) ));
				}
			}
		}
	}

//---------2.shadow map for light
	bias = 0.005*tan(acos(cosTheta));
	bias = clamp(bias, 0,0.003);
	//bias = -0.001;
	float lightVisibility = 1.0;

	if(lightShadowCoord.w >0.0)
	{
		
		vec4 lightShadowCoordProj = lightShadowCoord/lightShadowCoord.w;
		if(lightShadowCoordProj.x >= 0.0 && lightShadowCoordProj.x < 1.0 &&
			lightShadowCoordProj.y >= 0.0 && lightShadowCoordProj.y < 1.0 &&
			lightShadowCoordProj.z >= 0.0 && lightShadowCoordProj.z < 1.0)
		{
			if(textureProj( lightShadowMap, lightShadowCoord )  <  (lightShadowCoordProj.z-bias))
			{
				// Sample the shadow map 4 times
				for (int i=0;i<16;i++){
					//int index = i;
					int index = int(16.0*random(floor(pos_worldSpace.xyz*1000.0), i))%16;
					lightVisibility -= 0.0625*(1.0-texture(lightShadowMap, vec3(lightShadowCoordProj.xy + poissonDisk[index]/1000.0 , (lightShadowCoordProj.z-bias)) ));
				}
			}
		}
	}
//-----------------------------------------------end shadows
	
	fragColor = 
		vec4(// Ambient light, relative to window distance
		MaterialAmbientColor * 1/wind_dist +
		//global ambient
		MaterialAmbientColor / 2 +
		// Diffuse :
		lightVisibility * MaterialDiffuseColor  * lightPower * lightColor * cosTheta / (lightDistance*lightDistance) +
		// Specular :
		lightVisibility * MaterialSpecularColor * lightPower * pow(cosAlpha,5) / (lightDistance*lightDistance) +
		//fire:
		fireVisibility * MaterialFireColor * firePower * fireCosTheta / (fireDist*fireDist) +
		fireVisibility * MaterialFireSpecularColor * firePower * pow(fireCosAlpha,5) / (fireDist*fireDist)
		,1);

		//

	//check if fragment color is higher than a threshold and store this color
	float brightness = dot(fragColor.rgb, vec3(0.2126, 0.7152, 0.0722));
    if(brightness > 1.0){
        brightnessColor = vec4(fragColor.rgb, 1.0);
	}else{
		brightnessColor = vec4(0.0, 0.0, 0.0, 1.0);
	}


	////GOD RAYS//////////////
	
	if(cameraShadowCoord.w > 0.0 && lightShadowCoord.w > 0.0)
	{
		vec4 lightShadowCoordProj = lightShadowCoord/lightShadowCoord.w;
		vec4 cameraShadowCoordProj = cameraShadowCoord/cameraShadowCoord.w;
		
		if(lightShadowCoordProj.x >= 0.0 && lightShadowCoordProj.x < 1.0 &&
			lightShadowCoordProj.y >= 0.0 && lightShadowCoordProj.y < 1.0 &&
			lightShadowCoordProj.z >= 0.0 && lightShadowCoordProj.z < 1.0 &&
			cameraShadowCoordProj.x >= 0.0 && cameraShadowCoordProj.x < 1.0 &&
			cameraShadowCoordProj.y >= 0.0 && cameraShadowCoordProj.y < 1.0 &&
			cameraShadowCoordProj.z >= 0.0 && cameraShadowCoordProj.z < 1.0)
		{
			vec4 viewRay = lightShadowCoordProj - cameraShadowCoordProj;
			vec4 viewDir = normalize(viewRay);
			float rayLength = min(length(viewRay), 1);
			vec4 rayPoint = vec4(0,0,0,0);

			for(float marchedRay = 0.0f; marchedRay<=rayLength; marchedRay += stepSize)
			{
				rayPoint = cameraShadowCoordProj + marchedRay * viewDir;
			
				//rayPoint visible from light source?
				if(textureProj( lightShadowMap, rayPoint )  >  (rayPoint.z))
				{
					//add scattered light for god rays
					fragColor += vec4(lightColor,1) * 0.015f * pow((1- (rayPoint.z - 0.9f) * 16),1);
				}
				
			}
		}
		
	}
	
}